/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class org_gnunet_doodle_Doodle */

#ifndef _Included_org_gnunet_doodle_Doodle
#define _Included_org_gnunet_doodle_Doodle
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     org_gnunet_doodle_Doodle
 * Method:    getVersionInternal
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_gnunet_doodle_Doodle_getVersionInternal
  (JNIEnv *, jclass);

/*
 * Class:     org_gnunet_doodle_Doodle
 * Method:    open
 * Signature: (Ljava/lang/String;Lorg/gnunet/doodle/Doodle/Logger;)J
 */
JNIEXPORT jlong JNICALL Java_org_gnunet_doodle_Doodle_open
  (JNIEnv *, jclass, jstring, jobject);

/*
 * Class:     org_gnunet_doodle_Doodle
 * Method:    close
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_org_gnunet_doodle_Doodle_close
  (JNIEnv *, jclass, jlong);

/*
 * Class:     org_gnunet_doodle_Doodle
 * Method:    truncate
 * Signature: (JLjava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_org_gnunet_doodle_Doodle_truncate
  (JNIEnv *, jclass, jlong, jstring);

/*
 * Class:     org_gnunet_doodle_Doodle
 * Method:    expand
 * Signature: (JLjava/lang/String;Ljava/lang/String;)Z
 */
JNIEXPORT jboolean JNICALL Java_org_gnunet_doodle_Doodle_expand
  (JNIEnv *, jclass, jlong, jstring, jstring);

/*
 * Class:     org_gnunet_doodle_Doodle
 * Method:    lastModified
 * Signature: (JLjava/lang/String;)J
 */
JNIEXPORT jlong JNICALL Java_org_gnunet_doodle_Doodle_lastModified
  (JNIEnv *, jclass, jlong, jstring);

/*
 * Class:     org_gnunet_doodle_Doodle
 * Method:    search
 * Signature: (JLjava/lang/String;IZLjava/util/Vector;)I
 */
JNIEXPORT jint JNICALL Java_org_gnunet_doodle_Doodle_search
  (JNIEnv *, jclass, jlong, jstring, jint, jboolean, jobject);

#ifdef __cplusplus
}
#endif
#endif
